package org.processmining.primeminer.plugins;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

import javax.swing.JOptionPane;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XCostExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNetArray;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetArrayFactory;
import org.processmining.acceptingpetrinet.models.impl.AcceptingPetriNetFactory;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.primeminer.dialogs.YourDialog;


public class PrimeMiner 
{
	private static AcceptingPetriNetArrayFactory F;
	private static AcceptingPetriNetArray A;

	private static ArrayList<Tuple> independencies;
	private static XConceptExtension xconcept;
	private static XLifecycleExtension xlifecycle;
	private static XCostExtension xcost;
	private static ArrayList<Event> actualEvents;
	private static ArrayList<Event> coSetEvents;

	private static boolean debug = false;
	private static String oracle = "";
	private static int logSize = 0;
	private static int logSizeWithoutPrefixes = 0;
	private static int languageSize = 0;
	private static int[] languageDistribution;
	private static int partialLanguageSize = 0;
	private static int[] partialLanguageDistribution;
	
	private static ArrayList<ArrayList<String>> S;

	private static ArrayList<Integer> coSetCount = new ArrayList<Integer>();
	
	@Plugin(name = "Prime Miner", parameterLabels = {}, returnLabels = {
	"Prime Miner" }, returnTypes = {AcceptingPetriNetArray.class }, userAccessible = true, help = "reads event log and discovers workflow models.")
	@UITopiaVariant(affiliation = "FernUniversitaet in Hagen", author = "Robin Bergenthum", email = "robin.bergenthum@fernuni-hagen.de")
	
	
	public static AcceptingPetriNetArray runPrimeMiner(UIPluginContext context, XLog log)
	{
		String dialog = "";
		AcceptingPetriNetFactory I = new AcceptingPetriNetFactory();
		F = new AcceptingPetriNetArrayFactory();
		A = F.createAcceptingPetriNetArray();
		
		PrimeEventStructure pes;
		Petrinet petriNet = null;	
		
		ArrayList<String> oracleChoice;
		oracleChoice = calcOracle(log);

		S = calcLanguage(log);

		dialog=addDialogLn(dialog,  "                                                                             ");
		dialog=addDialogLn(dialog," ... log size               : " + logSize);
		dialog=addDialogLn(dialog," ... complete cases         : " + logSizeWithoutPrefixes);
		dialog=addDialogLn(dialog," ... language size          : " + languageSize);
		dialog=addDialogLn(dialog," ... possible oracles       : " + oracleChoice.toString());
		
		oracle = oracleChoice.get(0);
		YourDialog theD = new YourDialog(context, dialog,oracleChoice);
		JOptionPane.showMessageDialog(null, theD, "Prime Miner",JOptionPane.PLAIN_MESSAGE,null);
				
		oracle = oracleChoice.get(theD.list.getSelectedIndex());
		dialog=addDialogLn(dialog," ... oracle                 : " + oracle);
		
		
		HashMap<String, ArrayList<String>> C = null;
		if (oracle == "lifecycle") C = calcConcurrentLifecycle(S);
		if (oracle == "alpha") C = calcConcurrentAlpha(S);
		if (oracle == "none") C = calcConcurrentNone(S);
		
		String concs= "";
		ArrayList<String> lables = new ArrayList<String>();
		for (String s:C.keySet()) lables.add(s);
		for (int i = 0; i < lables.size()-1; i++)
		{
			for (int ii = i + 1; ii < lables.size(); ii++)
			{
				if (C.get(lables.get(i)).contains(lables.get(ii))) concs = concs + "(" + lables.get(i) +","+ lables.get(ii) + ") "; 
			}	
		}
		dialog=addDialogLn(dialog," ... concurrent actions     : " + concs);	
		
		//C.get("Restart Repair").add("Inform User");
		//C.get("Inform User").add("Restart Repair");
		
		ArrayList<Hasse> L = calcPartialLanguage(S,C);

		boolean[] go = new boolean[L.size()];
		int[] neuesNetzAnStelle = new int[L.size()];
		for(int index = 0; index < neuesNetzAnStelle.length;index ++) neuesNetzAnStelle[index] = Integer.MAX_VALUE;  
		
		int nrHasse=0;

		int platze = 0;
		int transitionen = 0;
		int arcs = 0;
				
		dialog=addDialogLn(dialog," ... partial language size  : " + partialLanguageSize);
		dialog=addDialog(dialog,  " ... cases per Hasse diagram: "); for(int nr:partialLanguageDistribution) dialog=addDialog(dialog,nr + " ");
		dialog=addDialogLn(dialog,  " ");			
		
		YourDialog theD2 = new YourDialog(context, dialog, partialLanguageSize);
		JOptionPane.showMessageDialog(null, theD2, "Prime Miner",JOptionPane.PLAIN_MESSAGE,null);
		int maxruns= theD2.sliderValue;
		boolean single = true;
		
		dialog=addDialogLn(dialog," ");
		
		if (theD2.list.getSelectedIndex()==1) single = false;
		
		if (debug) System.out.println("SINGLE :"+single);
		
		long start = System.currentTimeMillis( );
		boolean time = true;
	
		while (nrHasse < L.size() && nrHasse < maxruns)
		{
			
			if (System.currentTimeMillis( )-start > 30000) time = false;
			
			if (single) 
			{
				Hasse h = L.get(nrHasse);

				petriNet = buildPetriNet(context);
				Hashtable<Node, Transition> table = new Hashtable<Node, Transition>();
				for (Node n : h.events)
				{
					Transition t = petriNet.addTransition(n.name);
					table.put(n, t);
				}

				int anz = 0;

				for (Node n : h.events)
				{
					for (Node post : n.post)
					{
						Place p = petriNet.addPlace("p"+anz++);
						petriNet.addArc(p, table.get(post));
						petriNet.addArc(table.get(n),p);
					}
				}

				//for (int index = 0;index < L.size();index ++) go[index]=false;
				//go[nrHasse]=true;
				//Pes P = partialLanguageToPes(L, go);

				AcceptingPetriNet aPN = I.createAcceptingPetriNet(petriNet, new Marking(), new Marking());
				neuesNetzAnStelle[nrHasse] = A.getSize();
				A.addNet(aPN);
			}

			
			if (!single & time)
			{
				go[nrHasse]=true;
				System.out.println(" ... " + (nrHasse+1));
	
				Pes P = partialLanguageToPes(L, go);
				pes = P.getPrime();
				addStatesToPES(pes);
				
				petriNet = synthesize(context, pes);
			
				// Netz schon?
				ArrayList<Transition> tOhneVor = new ArrayList<Transition>();
				tOhneVor.addAll(petriNet.getTransitions());
				
				for(Transition t : petriNet.getTransitions())
				{
					for(Place p : petriNet.getPlaces())
					{
						if (petriNet.getArc(p, t) != null) tOhneVor.remove(t);
					}
				}
				
				if (tOhneVor.size()>0)
				{
					go[nrHasse] = false;
					neuesNetzAnStelle[nrHasse] = Integer.MAX_VALUE;
				}
				else
				{
					Marking initialMarking = new Marking();
					for (Place p:petriNet.getPlaces()) if(p.getLabel().equals("p1")) initialMarking.add(p);
					AcceptingPetriNet aPN = I.createAcceptingPetriNet(petriNet, initialMarking, new Marking());
					boolean neuesNetz=false;
					if (platze != petriNet.getPlaces().size()) neuesNetz=true;
					if (transitionen != petriNet.getTransitions().size()) neuesNetz=true;
					int anz=0;
					for (Place p:petriNet.getPlaces()) anz = anz + petriNet.getInEdges(p).size() + petriNet.getOutEdges(p).size();      		
					if (arcs != anz) neuesNetz=true;
					if (neuesNetz) 
					{
						A.addNet(aPN);
						platze = petriNet.getPlaces().size();
						transitionen = petriNet.getTransitions().size();
						arcs = anz;
					}
					neuesNetzAnStelle[nrHasse] = A.getSize();
				}
			}
			nrHasse++;
		}
		
		for (int index=0; index < A.getSize(); index++)
		{
			Petrinet pn = A.getNet(index).getNet();
			int gesamt = 0;
			int ausfuehrbar = 0;
			
			for (ArrayList<String> w: S)
			{
				boolean enabled = true;
				Marking m = new Marking();
				for (Place p:pn.getPlaces()) if(p.getLabel().equals("p1")) m.add(p);
				
				for (String lable : w)
				{
					lable = lable.trim();
					if (!lable.endsWith("start"))
					{
						lable = lable.split("_")[0];
						
						for (Transition t:pn.getTransitions())
						{
							if (t.getLabel().equals(lable))
							{
								if (!isEnabled(t, m, pn)) enabled = false;
								else fire(t, m, pn);
							}
						}
					}
				}
				if (enabled) ausfuehrbar = ausfuehrbar + languageDistribution[S.indexOf(w)];
				gesamt = gesamt + languageDistribution[S.indexOf(w)];
								
			}
			
			double prozent = ausfuehrbar * 100 / gesamt;

			if (!single) dialog=addDialog(dialog," ... Net "+ (index + 1) + " synthesized from Hasse diagrams");
			if (single) dialog=addDialog(dialog," ... Net "+ (index + 1) + " is the partial word");
			if (!single) for (int i = 0; i < partialLanguageSize; i++) if (neuesNetzAnStelle[i] <= (index + 1)) dialog=addDialog(dialog," " + (i + 1));
			if (single) for (int i = 0; i < partialLanguageSize; i++) if (neuesNetzAnStelle[i] == (index)) dialog=addDialogLn(dialog," " + (i + 1));
			if (!single) dialog=addDialogLn(dialog," can execute " + prozent + "% of the traces");
			
		}	
				
		// Print out DIALOG
		if (debug) System.out.println(dialog);
		if (!time) dialog=addDialog(dialog," ... time is up!");
		if (!single) {
		YourDialog theD3 = new YourDialog(context, dialog);
		JOptionPane.showMessageDialog(null, theD3, "Prime Miner",JOptionPane.PLAIN_MESSAGE,null);
		}
		return A;
	}
	
	public static boolean isEnabled(Transition t, Marking m, Petrinet pn)
	{
		boolean enabled = true;

		for (Place p:pn.getPlaces())
		{
			Arc a = pn.getArc(p,t);
			if (a!=null)
			{
				if (!m.contains(p)) enabled = false;
			}
		}
		return enabled;
	}
	
	public static void fire(Transition t, Marking m, Petrinet pn)
	{
		for (Place p:pn.getPlaces())
		{
			Arc a = pn.getArc(p,t);
			if (a!=null)
			{
				m.remove(p);
			}
		}

		for (Place p:pn.getPlaces())
		{
			Arc a = pn.getArc(t,p);
			if (a!=null)
			{
				m.add(p);
			}
		}
	}
	
	private static String addDialog(String s1, String s2) {return s1 + s2;}
	
	private static String addDialogLn(String s1, String s2) {return s1 + s2 +"\n";}


	public static String arrayToString(int[] array)
	{
		String arrayString = "";

		for (int i : array)
		{
			arrayString += i + ", ";
		}
		return arrayString;
	}

	// Prüft, ob eine Region ein Subset einer anderen Region ist
	public static boolean isSubset(Region subsetRegion, Region supersetRegion)
	{
		boolean isSubset = true;

		ArrayList<State> superElements = supersetRegion.elements;

		for (State state : subsetRegion.elements)
		{
			if (!superElements.contains(state))
			{
				isSubset = false;
			}
		}
		return isSubset;
	}

	// Prüft, ob eine Region ein Subset einer anderen Region ist
	public static boolean isEqual(Region region1, Region region2) {
		boolean isEqual = true;

		if (region1.elements.size() != region2.elements.size()) {
			return false;
		}

		ArrayList<State> elements1 = region1.elements;

		for (State state : region2.elements) {
			if (!elements1.contains(state)) {
				isEqual = false;
			}
		}
		return isEqual;
	}


	public static Transition getTransition(Petrinet petriNet, String label)
	{
		Transition transition = null;

		for (Transition t : petriNet.getTransitions())
		{
			if (t.getLabel().equals(label))
			{
				transition = t;
				break;
			}
		}
		return transition;
	}


	// Baut eine Menge von Regionen aus den Mengen von Switching- und Excitation-Regions
	// und vermeidet, dass es doppelte Regionen gibt
	public static ArrayList<Region> buildInitialRegions(ArrayList<Region> sw, ArrayList<Region> ex)
	{
		ArrayList<Region> regions = new ArrayList<Region>();
		ArrayList<String> regionStrings = new ArrayList<String>();

		for (Region r : sw)
		{
			regions.add(r);
			regionStrings.add(r.toString());
		}

		for (Region r : ex)
		{
			boolean vorhanden = false;

			for (String s : regionStrings)
			{
				if (s.equals(r.toString()))
				{
					vorhanden = true;
					break;
				}
			}

			if (!vorhanden)
			{
				regions.add(r);
				regionStrings.add(r.toString());
			}
		}

		return regions;
	}

	public static void deleteSubsets(ArrayList<Region> regions)
	{
		ArrayList<Region> removeRegions = new ArrayList<Region>();

		for (int i = 0; i < regions.size(); i++)
		{
			for (int k = 0; k < regions.size(); k++)
			{
				if (i != k)
				{
					if (isSubset(regions.get(i), regions.get(k)))
					{
						removeRegions.add(regions.get(i));
					}
				}

			}
		}

		regions.removeAll(removeRegions);
	}

	public static ArrayList<Region> calculateSwitchingRegions(PrimeEventStructure pes)
	{
		ArrayList<Region> allSwitchingRegions = new ArrayList<Region>();
		ArrayList<String> allLabels = getAllLabels(pes);


		// Über alle Labels iterieren
		for (String label : allLabels)
		{
			// Switching Regionen von label berechnen
			ArrayList<Region> switchingRegionsForLabel = new ArrayList<Region>();

			Region emptyRegion = new Region();
			switchingRegionsForLabel.add(emptyRegion);

			// Alle Events mit diesem Label holen
			ArrayList<Event> events = getAllEventsWithLabel(pes, label);

			// Über alle Events mit gleichem Label iterieren und Menge an CoSets holen
			ArrayList<Integer> coSets = new ArrayList<Integer>();

			for (Event event : events)
			{

				// Über Co-Sets von event iterieren
				for (int coSet : event.coSets)
				{
					// Speichern von allen Events im Nachbereich, die im gleichen CoSet sind
					ArrayList<State> nextStatesInCoSet = new ArrayList<State>();

					for (State nextState : event.nextStates)
					{
						if (nextState.outEvent.coSets.contains(coSet)) nextStatesInCoSet.add(nextState);
					}


					ArrayList<Region> newRegions = new ArrayList<Region>();
					ArrayList<Region> zuLoeschendeRegionen = new ArrayList<Region>();

					// Über alle bisherigen Regionen iterieren
					for (Region r : switchingRegionsForLabel)
					{
						// Wenn von den nachfolgenden States des aktuellen Events, im aktuellen CoSet
						// der aktuellen Region noch keiner ausgewählt wurde, dann kann man weitermachen
						// Wenn schon einer ausgewählt wurde, dann wird das Event in diesem CoSet in dieser
						// Region einfach übersprungen (da muss nichts kombiniert werden)
						if (nochKeinNextStateSelected(event, coSet, r))
						{
							boolean loeschen = false;

							for (State nextState : nextStatesInCoSet)
							{
								if (nextStatesInCoSet.size() == 1)
								{
									// Nur einen Nachfolger -> Einfach zur aktuellen Region hinzufügen
									if (!r.elements.contains(nextState))
									{
										r.elements.add(nextState);
										newRegions.add(r);
									}
								}
								else
								{
									// Mehrere Nachfolger in gleichem CoSet -> Kombinationen bilden

									Region newRegion = new Region();
									for (State s : r.elements) newRegion.elements.add(s);
									newRegion.elements.add(nextState);
									newRegions.add(newRegion);
									// Aktuellen States in diesem CoSet als ausgewählt markieren
									nextState.selected.put(coSet, true);

									// Alle anderen States in diesem CoSet als markiert deklarieren
									for (State s : nextStatesInCoSet)
										s.markiert.put(coSet, true);

									loeschen = true;
								}
							}

							if (loeschen) zuLoeschendeRegionen.add(r);
						}

					}

					for (Region r : zuLoeschendeRegionen)
					{
						switchingRegionsForLabel.remove(r);
					}

					switchingRegionsForLabel.addAll(newRegions);
					newRegions.clear();
				}
			}

			allSwitchingRegions.addAll(switchingRegionsForLabel);

		}

		ArrayList<Region> switchingRegions = new ArrayList<Region>();

		for (Region r : allSwitchingRegions)
		{
			if (r.elements.size() != 0)
				switchingRegions.add(r);
		}


		ArrayList<Region> reducedSwitchingRegions = new ArrayList<Region>();

		for (Region r : switchingRegions)
		{
			if (!reducedSwitchingRegions.contains(r)) reducedSwitchingRegions.add(r);
		}

		return reducedSwitchingRegions;
	}

	public static boolean schnittIstNichtLeer(ArrayList<Integer> liste1, ArrayList<Integer> liste2)
	{
		boolean istNichtLeer = false;

		for (int i = 0; i < liste1.size(); i++)
		{
			if (liste2.contains(liste1.get(i))) istNichtLeer = true;
		}

		return istNichtLeer;
	}

	public static ArrayList<Region> calculateExcitationRegions(PrimeEventStructure pes)
	{
		ArrayList<Region> allExcitationRegions = new ArrayList<Region>();
		ArrayList<String> allLabels = getAllLabels(pes);


		// Über alle Labels iterieren
		for (String label : allLabels)
		{
			// Excitation Regionen von label berechnen
			ArrayList<Region> excitationRegionsForLabel = new ArrayList<Region>();

			Region emptyRegion = new Region();
			excitationRegionsForLabel.add(emptyRegion);

			// Alle Events mit diesem Label holen
			ArrayList<Event> events = getAllEventsWithLabel(pes, label);


			// Über alle Events mit gleichem Label iterieren und Menge an CoSets holen
			ArrayList<Integer> coSets = new ArrayList<Integer>();

			for (Event event : events)
			{


				// Über Co-Sets von event iterieren
				for (int coSet : event.coSets)
				{
					// Speichern von allen Events im Vorbereich, die im gleichen CoSet sind
					ArrayList<State> previousStatesInCoSet = new ArrayList<State>();

					for (State previousState : event.previousStates)
					{
						if (previousState.inEvent.coSets.contains(coSet)) previousStatesInCoSet.add(previousState);
					}


					ArrayList<Region> newRegions = new ArrayList<Region>();
					ArrayList<Region> zuLoeschendeRegionen = new ArrayList<Region>();

					// Über alle bisherigen Regionen iterieren
					for (Region r : excitationRegionsForLabel)
					{
						// Wenn von den nachfolgenden States des aktuellen Events, im aktuellen CoSet
						// der aktuellen Region noch keiner ausgewählt wurde, dann kann man weitermachen
						// Wenn schon einer ausgewählt wurde, dann wird das Event in diesem CoSet in dieser
						// Region einfach übersprungen (da muss nichts kombiniert werden)
						if (nochKeinPreviousStateSelected(event, coSet, r))
						{
							boolean loeschen = false;

							for (State previousState : previousStatesInCoSet)
							{
								if (previousStatesInCoSet.size() == 1)
								{
									// Nur einen Nachfolger -> Einfach zur aktuellen Region hinzufügen
									if (!r.elements.contains(previousState))
									{
										r.elements.add(previousState);
										newRegions.add(r);
									}
								}
								else
								{
									// Mehrere Vorgänger in gleichem CoSet -> Kombinationen bilden

									Region newRegion = new Region();
									for (State s : r.elements) newRegion.elements.add(s);
									newRegion.elements.add(previousState);
									newRegions.add(newRegion);
									// Aktuellen States in diesem CoSet als ausgewählt markieren
									previousState.selected.put(coSet, true);

									// Alle anderen States in diesem CoSet als markiert deklarieren
									for (State s : previousStatesInCoSet)
										s.markiert.put(coSet, true);

									loeschen = true;
								}
							}

							if (loeschen) zuLoeschendeRegionen.add(r);
						}

					}

					for (Region r : zuLoeschendeRegionen)
					{
						excitationRegionsForLabel.remove(r);
					}

					excitationRegionsForLabel.addAll(newRegions);
					newRegions.clear();
				}
			}

			allExcitationRegions.addAll(excitationRegionsForLabel);

		}

		ArrayList<Region> excitationRegions = new ArrayList<Region>();

		for (Region r : allExcitationRegions)
		{
			if (r.elements.size() != 0)
				excitationRegions.add(r);
		}


		ArrayList<Region> reducedExcitationRegions = new ArrayList<Region>();

		for (Region r : excitationRegions)
		{
			if (!reducedExcitationRegions.contains(r)) reducedExcitationRegions.add(r);
		}

		return reducedExcitationRegions;
	}

	public static void addStatesToPES(PrimeEventStructure pes)
	{
		// Über alle Events iterieren und States hinzufügen
		for (Event event : pes.events)
		{
			for (Event nextEvent : event.nextEvents)
			{
				State state = new State();

				state.inEvent = event;
				state.outEvent = nextEvent;

				event.nextStates.add(state);
				nextEvent.previousStates.add(state);

				// CoSets zum State hinzufügen
				for (Integer coSet : event.coSets)
				{
					state.markiert.put(coSet, false);
				}

				pes.states.add(state);
			}
		}
	}

	public static ArrayList<String> getAllLabels(PrimeEventStructure pes)
	{
		ArrayList<String> labels = new ArrayList<String>();

		for (Event event : pes.events)
		{
			if (!labels.contains(event.eventName))
				labels.add(event.eventName);
		}

		return labels;
	}

	public static ArrayList<Event> getAllEventsWithLabel(PrimeEventStructure pes, String label)
	{
		ArrayList<Event> events = new ArrayList<Event>();

		for (Event event : pes.events)
		{
			if (event.eventName.equals(label))
				events.add(event);
		}

		return events;
	}

	public static void addInitialAndFinalEvents(PrimeEventStructure pes)
	{
		// Für jedes Event ohne ausgehende Kanten ein finales Event erzeugen
		// Initiales StartEvent gibt es bereits
		ArrayList<Event> finalEvents = new ArrayList<Event>();

		int finalEventIndex = 1;

		for (Event e : pes.events)
		{
			if (e.nextEvents.size() == 0)
			{
				Event finalEvent = new Event("finalEvent" + finalEventIndex);

				finalEvent.previousEvents.add(e);
				finalEvent.coSets.addAll(e.coSets);
				e.nextEvents.add(finalEvent);
				finalEvents.add(finalEvent);

				finalEventIndex++;
			}
		}

		pes.events.addAll(finalEvents);
	}

	public static Petrinet buildPetriNet(PluginContext context)
	{
		Petrinet petriNet = PetrinetFactory.newPetrinet("Test");
		//Marking marking = new Marking();
		//context.addConnection(new InitialMarkingConnection(petriNet, marking));

		//return new Object[] {petriNet, marking};
		return petriNet;
	}

	public static ArrayList<String> calcOracle(XLog log)
	{	
		ArrayList<String> oracleChoice = new ArrayList<String>();
		
		xconcept = XConceptExtension.instance();
		xlifecycle = XLifecycleExtension.instance();
		xcost = XCostExtension.instance();

		coSetCount = new ArrayList<Integer>();
		actualEvents = new ArrayList<Event>();
		coSetEvents = new ArrayList<Event>();
		independencies = new ArrayList<Tuple>();

		boolean cost = true;
		boolean life = true;
		boolean lifeStart = false;
		boolean lifeComplete = false;

		for (int t = 0; t < log.size(); t++)
		{
			XTrace trace = log.get(t);

			for (int e = 0; e < trace.size(); e++)
			{
				XEvent event = trace.get(e);
				String costy = xcost.extractCurrency(event);
				String lifecycle = xlifecycle.extractTransition(event);

				if (lifecycle != null)
				{
					if (lifecycle.equals("start")) lifeStart = true;
					if (lifecycle.equals("complete")) lifeComplete = true;
				}

				if (costy == null) cost = false;
				if (lifecycle == null) life = false;
			}
		}

		if (cost) oracleChoice.add("later-than");
		if (life) if (lifeStart && lifeComplete) oracleChoice.add("lifecycle");
		oracleChoice.add("alpha");
		oracleChoice.add("none");
		return oracleChoice;
	}	
	
	
	public static PrimeEventStructure loadPES(XLog log) {
		// ======================================================================
		// Abhängigkeitsmatrix bauen
		// ======================================================================


		// Number of Events holen und Traces mit mehrfachen Events löschen (Nur für die Berechnung der Unabhängigkeiten)
		ArrayList<String> labels = new ArrayList<String>();
		ArrayList<XTrace> logTraces = new ArrayList<XTrace>();

		for (int t = 0; t < log.size(); t++)
		{
			XTrace trace = log.get(t);
			logTraces.add(trace);

			for (int e = 0; e < trace.size(); e++)
			{
				XEvent eventA = trace.get(e);

				String eventNameA = xconcept.extractName(eventA);

				if (!labels.contains(eventNameA))
					labels.add(eventNameA);
			}
		}

		int numberOfLabels = labels.size();


		Map<String, Event> labelMap = new HashMap<String, Event>();

		for (String label : labels)
		{
			labelMap.put(label, null);
		}


		// =====================================================================
		// Prime Event Structure bauen

		PrimeEventStructure pes = new PrimeEventStructure();

		Event startEvent = new Event("initialEvent");
		pes.events.add(startEvent);
		coSetEvents = new ArrayList<Event>(); // Menge der Events im aktuellen CoSet

		int coSetIndex = 0; // Der Index des aktuellen CoSets;
		int letztesCoSet = 0;

		for (int t = 0; t < log.size(); t++)
		{
			XTrace trace = log.get(t);

			for (String label : labels)
			{
				labelMap.put(label, null);
			}

			// Wurde ein neues Event gefunden? Wenn nicht, dann muss kein
			// neues CoSet erzeugt werden
			boolean neuesCoSetErstellen = false;

			// Aktuelles CoSet zurücksetzen
			coSetEvents.clear();
			coSetEvents.add(startEvent);

			for (int e = 0; e < trace.size(); e++)
			{
				// Event auslesen
				XEvent event = trace.get(e);
				String eventName = xconcept.extractName(event);
				String lifecycle = xlifecycle.extractTransition(event);
				String cost = xcost.extractCurrency(event);

				String[] preEvents = cost.split(",");

				if (lifecycle != null)
					eventName = eventName + lifecycle;





				for (String preEvent : preEvents)
				{
					if (preEvent.equals(""))
					{
						boolean found = false;

						for (Event ev : startEvent.nextEvents)
						{
							if (ev.eventName.equals(eventName))
							{
								found = true;
								labelMap.put(eventName, ev);
							}
						}

						if (!found)
							neuesCoSetErstellen = true;
					}
					else
					{
						boolean found = false;

						for (Event ev : labelMap.get(preEvent).nextEvents)
						{
							if (ev.eventName.equals(eventName))
							{
								found = true;
								labelMap.put(eventName, ev);
							}
						}

						if (!found)
							neuesCoSetErstellen = true;
					}
				}

				if (!neuesCoSetErstellen)
				{
					coSetEvents.add(labelMap.get(eventName));


				}
				else
				{
					Event pesEvent = new Event(eventName);
					labelMap.put(eventName, pesEvent);
					pes.events.add(pesEvent);
					coSetEvents.add(pesEvent);

					for (String preEvent : preEvents)
					{
						if (preEvent.equals(""))
						{
							startEvent.nextEvents.add(pesEvent);
							pesEvent.previousEvents.add(startEvent);
						}
						else
						{
							labelMap.get(preEvent).nextEvents.add(pesEvent);
							pesEvent.previousEvents.add(labelMap.get(preEvent));
						}
					}
				}

			}

			// Neues CoSet für den aktuellen Trace soll gebaut werden
			if (neuesCoSetErstellen) {
				coSetIndex++;

				coSetCount.add(1);

				// Jedem Event des CoSets wird seine Index-Nummer hinzugefügt
				for (Event ee : coSetEvents) {
					ee.coSets.add(coSetIndex);
				}

				//                    // Das neue CoSet bzw. die Menge der Events des CoSets der PES hinzufügen
				//                    // Ein Event wird nur hinzugefügt, wenn es noch nicht vorhanden ist
				//                    for (Event ee : coSetEvents) {
				//                        if (!pes.events.contains(ee))
				//                            pes.events.add(ee);
				//                    }
			}
			else
			{
				// Keine neue LPO wurde erstellt. Heißt sie ist schon in der PES
				// vorhanden. Zähler für diese LPO hochzählen (für dieses CoSet)
				coSetCount.set(coSetIndex-1, coSetCount.get(coSetIndex-1) + 1);
			}

		}

		return pes;
	}


	public static ArrayList<ArrayList<String>> calcLanguage(XLog log) 
	{
		// Language fur die Worter 
		ArrayList<ArrayList<String>> language = new ArrayList<ArrayList<String>>();
		// Die Traces pro Wort zahlen
		ArrayList<Integer> count = new ArrayList<Integer>();
		// Labels
		ArrayList<String> lables = new ArrayList<String>();

		logSize = log.size();
		if (debug) System.out.println(" ... log size               : " + logSize);


		// Log laden und Sprache bilden
		for (int t = 0; t < log.size(); t++)
		{
			XTrace trace = log.get(t);
			ArrayList<String> nextTrace = new ArrayList<String>();

			for (int e = 0; e < trace.size(); e++)
			{
				XEvent event = trace.get(e);
				// OHNE END
				//if (!xconcept.extractName(event).equals("end"))
				//{
					nextTrace.add(xconcept.extractName(event)+"_"+ xlifecycle.extractTransition(event) + " ");
					if (debug) System.out.print(xconcept.extractName(event)+"_"+ xlifecycle.extractTransition(event)+" ");
					if (!lables.contains(xconcept.extractName(event))) lables.add(xconcept.extractName(event));
				//}
			}
			if (debug) System.out.println();

			boolean found = false;
			ArrayList<String> a;
			for (int i=0;i<language.size();i++)
			{
				a=language.get(i);
				boolean passt = true; 
				if (a.size() == nextTrace.size())
				{
					for (int ii=0; ii < a.size();ii++) 
					{
						if (!a.get(ii).equals(nextTrace.get(ii))) passt = false; 
					}
				}
				else 
				{
					passt=false; 
				}
				if (passt) 
				{
					count.set(i,count.get(i)+1);
					found = true;
				}
			}  
			if (!found) 
			{
				language.add(nextTrace);
				count.add(1);
			}
		}

		// prefixe loschen
		int geloescht = 0;
		boolean[] pre= new boolean[language.size()];
		for (ArrayList<String> w1 : language)
		{
			for (ArrayList<String> w2 : language)
			{
				if (write(w1).startsWith(write(w2)) && !w1.equals(w2)) pre[language.indexOf(w2)] = true;
			}
		}

		for (int i = language.size()-1;i>=0;i--)
		{
			if (pre[i])
			{
				geloescht = geloescht + count.get(i);
				count.remove(i);
				language.remove(i);
			}
		}		   

		// Sprache nach Anzahl der Worter sortieren
		ArrayList<String> wort;
		Integer anzahl;
		Integer max=0;

		for (int i=0;i<language.size();i++)
		{
			anzahl = count.get(i);
			wort = language.get(i);
			max=i;
			for (int ii=i; ii < language.size();ii++) 
			{
				if (count.get(ii) > anzahl)
				{
					anzahl = count.get(ii);
					wort = language.get(ii);
					max=ii;
				}
			}
			count.set(max,count.get(i));
			count.set(i, anzahl);
			language.set(max,language.get(i));
			language.set(i, wort);
		} 

		// STATS !!!
		languageSize = language.size();

		logSizeWithoutPrefixes = logSize - geloescht;
		if (debug)  System.out.println(" ... # complete cases       : " + logSizeWithoutPrefixes);
		if (debug)  System.out.println(" ... language size          : " + languageSize);

		languageDistribution = new int[languageSize];
		for (int index = 0; index < languageSize; index++)
		{
			languageDistribution[index] = count.get(index);
		}

		return language;
	}

	//
	//  calcConcurrentAlpha
	//
	public static HashMap<String, ArrayList<String>> calcConcurrentAlpha(ArrayList<ArrayList<String>> language) 
	{
		HashMap<String, ArrayList<String>> concurrent = new HashMap<String, ArrayList<String>>();
		ArrayList<String> lables = new ArrayList<String>();

		ArrayList<ArrayList<String>> languageOnlyComplete = new ArrayList<ArrayList<String>>();
		
		for (ArrayList<String> wort : language) 
		{
			ArrayList<String> neuWort= new ArrayList<String>();
			languageOnlyComplete.add(neuWort);
			for (String s : wort) if (!s.trim().endsWith("start")) neuWort.add(s.split("_")[0]);
		}
		
		for (ArrayList<String> wort : languageOnlyComplete) for (String s : wort) if (!lables.contains(s)) lables.add(s);

		int[][] dependencyMatrix= new int[lables.size()][lables.size()];		
		
		for (ArrayList<String> wort : languageOnlyComplete)
		{
			for (int index = 0; index < wort.size() - 1; index++)
			{
				dependencyMatrix[lables.indexOf(wort.get(index))][lables.indexOf(wort.get(index+1))]++;
			}
		}
		
		int max=0;
		for (int[] line:dependencyMatrix) for(int x:line) if(x>max) max = x;
		
		for (String l:lables)
		{
			ArrayList<String> list = new ArrayList<String>();
			concurrent.put(l.split("_")[0], list);
			
			for (String l2:lables)
			{
				if (!l.equals(l2))
				{
					if (dependencyMatrix[lables.indexOf(l)][lables.indexOf(l2)] > 0 + Math.floor(Math.log10(max)))
					{
						if (dependencyMatrix[lables.indexOf(l2)][lables.indexOf(l)] > 0 + Math.floor(Math.log10(max)))
						{
							list.add(l2.split("_")[0]);
						}
					}
				}
			}
			
		}
		
		return concurrent;
	}
	
	//
	//  calcConcurrentNone
	//
	public static HashMap<String, ArrayList<String>> calcConcurrentNone(ArrayList<ArrayList<String>> language) 
	{
		HashMap<String, ArrayList<String>> concurrent = new HashMap<String, ArrayList<String>>();
		ArrayList<String> lables = new ArrayList<String>();

		for (ArrayList<String> wort : language) for (String s : wort) if (!lables.contains(s)) lables.add(s);
		
		for (String l:lables)
		{
			ArrayList<String> list = new ArrayList<String>();
			concurrent.put(l.split("_")[0], list);
		}
		
		return concurrent;
	}
	
	public static HashMap<String, ArrayList<String>> calcConcurrentLifecycle(ArrayList<ArrayList<String>> language) 
	{

		// Hashmap mit Abhanigkeiten berechnen
		HashMap<String, ArrayList<String>> concurrent = new HashMap<String, ArrayList<String>>();
		HashMap<String, Boolean> offen = new HashMap<String, Boolean>();

		ArrayList<String> lables = new ArrayList<String>();
		for (ArrayList<String> wort : language) for (String s : wort) if (!lables.contains(s.split("_")[0])) lables.add(s.split("_")[0]);

		for (String s:lables) concurrent.put(s, new ArrayList<String>());
		for (String s:lables) offen.put(s, false);

		int anzahlTracesBetrachtet = 0;
		for (ArrayList<String> w:language)
		{

			// wieviele Traces werden fur die Con Matricx durchsucht ... hier 2/3 Prozent
			if (anzahlTracesBetrachtet<3*logSize/4)
			{	   
				for (String e: w)
				{
					e = e.trim();
					for(String l:lables)
					{
						if (offen.get(l)) 
						{
							if (!concurrent.get(e.split("_")[0]).contains(l.split("_")[0])) concurrent.get(e.split("_")[0]).add(l.split("_")[0]);
							if (!concurrent.get(l.split("_")[0]).contains(e.split("_")[0])) concurrent.get(l.split("_")[0]).add(e.split("_")[0]);
						}
					}

					if (e.endsWith("complete")) offen.put(e.split("_")[0], false);
					if (e.endsWith("start")) 
					{
						offen.put(e.split("_")[0], true);
					}

				}
				anzahlTracesBetrachtet = anzahlTracesBetrachtet + languageDistribution[language.indexOf(w)];
			}
		} 

		for (String name: lables){concurrent.get(name).remove(name);}

		if (debug) System.out.println("Con Matrix: " + concurrent.toString());
		if (debug) System.out.println("Worter gescannt: " + anzahlTracesBetrachtet);

		return concurrent;
	}

	public static ArrayList<Hasse> calcPartialLanguage(ArrayList<ArrayList<String>> language, HashMap<String, ArrayList<String>> concurrent) 
	{

		ArrayList<Hasse> partialLanguage = new ArrayList<Hasse>();

		for (ArrayList<String> w:language)
		{
			Hasse h = new Hasse(w,concurrent,languageDistribution[language.indexOf(w)]);

			boolean neu = true;
			for(Hasse hase:partialLanguage) if (hase.isTheSame(h)) 
			{
				neu = false;
				hase.anz= hase.anz + h.anz;
			}

			if (neu) partialLanguage.add(h);
		}

		Hasse park;
		for(int i = 0;i<partialLanguage.size()-1;i++)
		{
			for(int ii = i;ii < partialLanguage.size();ii++)
			{
				if (partialLanguage.get(ii).anz > partialLanguage.get(i).anz)
				{
					park = partialLanguage.get(i);
					partialLanguage.set(i,partialLanguage.get(ii));
					partialLanguage.set(ii,park);
				}
			}   
		}

		/// STATS !!!!
		partialLanguageSize = partialLanguage.size();
		if (debug) System.out.println(" ... partial Language size  : " + partialLanguageSize);
		partialLanguageDistribution = new int[partialLanguageSize];
		for (Hasse h : partialLanguage) partialLanguageDistribution[partialLanguage.indexOf(h)]=h.anz;  

		if (debug) for (Hasse hase:partialLanguage) hase.prt();

		return partialLanguage;
	}



	static Pes partialLanguageToPes(ArrayList<Hasse> partialLanguage, boolean[] go)
	{
		Pes pes = new Pes();
		HashMap<Node,PesNode> nodeTable = new HashMap<Node,PesNode>();

		PesNode init = new PesNode("init");
		pes.nodes.add(init);


		for (int i = 0; i < partialLanguage.size(); i++)
		{
			if (go[i])
			{
				init.coset.add(i);
				Hasse h = partialLanguage.get(i);
				for (Node n : h.events)
				{
					PesNode p = new PesNode(n.name);
					nodeTable.put(n,p);
					pes.nodes.add(p);
					p.coset.add(i);
				}

				for (Node n : h.events)
				{
					if (n.pre.isEmpty())
					{
						pes.addArc(init, nodeTable.get(n));
					}

					for(Node n2 : n.pre)
					{
						pes.addArc(nodeTable.get(n2), nodeTable.get(n));
					} 

				}
			}
		}

		// Prefixe verschmelzen
		pes.crop();
		if (debug) pes.draw();

		return pes;
	}

	public static Petrinet synthesize(PluginContext context, PrimeEventStructure pes)
	{
		Petrinet petriNet;
		ArrayList<Region> switchingRegions = calculateSwitchingRegions(pes);
		ArrayList<Region> excitationRegions = calculateExcitationRegions(pes);
		ArrayList<Region> regions = buildInitialRegions(switchingRegions, excitationRegions);
		ArrayList<int[]> gradients = new ArrayList<int[]>();
		ArrayList<Region> goodRegions = new ArrayList<Region>();
		ArrayList<int[]> goodGradients = new ArrayList<int[]>();

		// BadRegions reparieren
		while (!regions.isEmpty()) 
		{
			Region region = regions.remove(0);

			// Is Region ???
			boolean isRegion = true;
			boolean komplettRaus = false;

			Event badMinE = null;
			Event badMaxE = null;
			Integer badMinCo = null;
			Integer badMaxCo = null;

			int[] gradient = new int[getAllLabels(pes).size()];
			gradients.add(gradient);

			for (String label : getAllLabels(pes)) 
			{
				ArrayList<Event> events = getAllEventsWithLabel(pes, label);
				boolean firstTime = true;
				int g = 0;
				int minG = 0;
				int maxG = 0;

				for (Event event : events) 
				{
					if (!isRegion) break;
					for (int coSet : event.coSets) 
					{
						g = 0;

						// Speichern von allen Events im Nachbereich, die im gleichen CoSet sind
						for (State nextState : event.nextStates) {
							if (nextState.outEvent.coSets.contains(coSet) && region.elements.contains(nextState)) {
								g++;
							}
						}

						if (g > 1) {
							isRegion = false;
							komplettRaus = true;
						}

						// Speichern von allen Events im Vorbereich, die im gleichen CoSet sind
						for (State previousState : event.previousStates) {
							if (previousState.inEvent.coSets.contains(coSet)
									&& region.elements.contains(previousState)) {
								g--;
							}
						}

						if (g < -1) {
							isRegion = false;
							komplettRaus = true;
						}

						if (firstTime) {
							gradient[getAllLabels(pes).indexOf(label)] = g;
							firstTime = false;

							minG = g;
							maxG = g;

							badMinCo = coSet;
							badMaxCo = coSet;

							badMinE = event;
							badMaxE = event;
						} else {
							if (g != gradient[getAllLabels(pes).indexOf(label)]) {
								isRegion = false;

								if (g > maxG) {
									maxG = g;
									badMaxE = event;
									badMaxCo = coSet;
								}

								if (g < minG) {
									minG = g;
									badMinE = event;
									badMinCo = coSet;
								}
							}
						}
					}
				}
			}

			if (isRegion) 
			{
				goodRegions.add(region);
				goodGradients.add(gradient);
			} 
			else 
			{
				if (!komplettRaus) 
				{
					// Reparatur nach Min
					// Speichern von allen Events im Nachbereich, die im gleichen CoSet sind
					ArrayList<State> nextStatesInCoSet = new ArrayList<State>();

					for (State nextState : badMinE.nextStates) 
					{
						if (nextState.outEvent.coSets.contains(badMinCo))
							nextStatesInCoSet.add(nextState);
					}

					// Für jeden nachfolgenden State (Kante), neue Region erzeugen und zur Menge
					// der BadRegionen hinzufügen
					for (State s : nextStatesInCoSet) {
						if (!region.elements.contains(s)) {
							Region newRegion = new Region();
							for (State p : region.elements) {
								newRegion.elements.add(p);
							}
							newRegion.elements.add(s);

							boolean hinzu = true;
							for (Region goodRegion : goodRegions) {
								if (isSubset(goodRegion, newRegion))
									hinzu = false;
							}

							for (Region reg : regions) {
								if (isEqual(reg, newRegion))
									hinzu = false;
							}

							if (hinzu)
								regions.add(newRegion);
						}
					}

					// Reparatur nach Max
					ArrayList<State> previousStatesInCoSet = new ArrayList<State>();

					for (State previousState : badMaxE.previousStates) {
						if (previousState.inEvent.coSets.contains(badMaxCo))
							previousStatesInCoSet.add(previousState);
					}

					// Für jeden vorangehende State (Kante), neue Region erzeugen und zur Menge
					// der BadRegionen hinzufügen
					for (State s : previousStatesInCoSet) {
						if (!region.elements.contains(s)) {
							Region newRegion = new Region();
							for (State p : region.elements) {
								newRegion.elements.add(p);
							}
							newRegion.elements.add(s);

							boolean hinzu = true;
							for (Region goodRegion : goodRegions) {
								if (isSubset(goodRegion, newRegion))
									hinzu = false;
							}

							for (Region reg : regions) {
								if (isEqual(reg, newRegion))
									hinzu = false;
							}

							if (hinzu)
								regions.add(newRegion);
						}
					}
				}
			}
		}

		// Petrinetz aus Regionen bauen
		petriNet = buildPetriNet(context);

		for (String label : getAllLabels(pes))
		{
			if (!label.startsWith("initialEvent") && !label.startsWith("finalEvent"))
			{
				Transition newTransition = petriNet.addTransition(label.split("_")[0]);
			}
		}

		// Netz schoner machen
		ArrayList<int[]> removeGradientsList = new ArrayList<int[]>();
		ArrayList<Region> removeRegionsList = new ArrayList<Region>();

		for (int i = 0; i < goodGradients.size(); i++)
		{
			boolean weg = true;
			for (int k = 0; k < goodGradients.get(i).length; k++)
			{
				if ((goodGradients.get(i)[k] == -1) && (!getAllLabels(pes).get(k).startsWith("finalEvent"))) weg = false;
			}

			if (weg)
			{
				removeGradientsList.add(goodGradients.get(i));
				removeRegionsList.add(goodRegions.get(i));
			}
		}

		// Gleiche Gradienten rauswerfen
		for (int i = 0; i < goodGradients.size() - 1; i++)
		{
			for (int k = i + 1; k < goodGradients.size(); k++)
			{

				if (arrayToString(goodGradients.get(i)).equals(arrayToString(goodGradients.get(k))))
				{
					removeGradientsList.add(goodGradients.get(k));
					removeRegionsList.add(goodRegions.get(k));
				}
			}
		}

		goodGradients.removeAll(removeGradientsList);
		goodRegions.removeAll(removeRegionsList);

		ArrayList<String> nichtVerbundeneTransitionen = getAllLabels(pes);

		// Plaetze v
		int placeIndex = 1;
		for (int[] g : goodGradients) {
			Place newPlace = petriNet.addPlace("p" + placeIndex);
			placeIndex++;

			for (String label : getAllLabels(pes)) {
				if (g[getAllLabels(pes).indexOf(label)] == 1)
				{
					if (!label.equals("initialEvent"))
					{
						org.processmining.models.graphbased.directed.petrinet.elements.Arc newArc = petriNet.addArc(getTransition(petriNet, label), newPlace);
					}
				}
				else if (g[getAllLabels(pes).indexOf(label)] == -1)
				{
					if (!label.startsWith("finalEvent"))
					{
						org.processmining.models.graphbased.directed.petrinet.elements.Arc newArc = petriNet.addArc(newPlace, getTransition(petriNet, label));
						nichtVerbundeneTransitionen.remove(label);
					}
				}
			}
		}
		nichtVerbundeneTransitionen.remove("initialEvent");

		// Finale Plaetze konstruieren
		for(Transition t : petriNet.getTransitions())
		{
			if (t.getVisibleSuccessors().isEmpty()) 
			{
				Place p = petriNet.addPlace("p" + placeIndex);
				placeIndex++;
				petriNet.addArc(t, p);
			}
		}
		return petriNet;
	}
	
	
	static String write(ArrayList<String> w)
	{
		String s = "";
		for (String s1 : w) s = s + s1;
		return s;
	}


	public static PrimeEventStructure generatePES(XLog log) 
	{
		// ======================================================================
		// Abhängigkeitsmatrix bauen
		// ======================================================================

		// 1=unrelated, 2=parallel, 3=causal right, 4=causal left
		int[][] dependencyMatrix;
		int matrixIndex = 0;

		// Map, mapping event names to matrix positions
		Map<String, Integer> eventMap = new HashMap<String, Integer>();
		Map<Integer, String> eventMapReverse = new HashMap<Integer, String>();

		// Dependency Matrix initialisieren

		// Number of Events holen und Traces mit mehrfachen Events löschen (Nur für die Berechnung der Unabhängigkeiten)
		ArrayList<String> labels = new ArrayList<String>();
		ArrayList<XTrace> logTraces = new ArrayList<XTrace>();

		for (int t = 0; t < log.size(); t++)
		{
			XTrace trace = log.get(t);
			logTraces.add(trace);

			for (int e = 0; e < trace.size(); e++)
			{
				XEvent eventA = trace.get(e);
				String eventNameA = xconcept.extractName(eventA);

				if (!labels.contains(eventNameA))
					labels.add(eventNameA);
			}
		}

		independencies = new ArrayList<Tuple>();

		// Unabhängigkeiten berechnen auf Grundlage von Start und Complete Lifecycle
		// Über alle Label
		for (String label : labels)
		{
			// Über alle Traces
			for (XTrace trace : logTraces)
			{
				boolean startFound = false;
				boolean selfIndep = false;

				// Über alles Events des Trace
				for (int e = 0; e < trace.size() - 1; e++)
				{
					// Wenn label-Start vorkommt, dann alle nachfolgenden Events als
					// unabhängig deklarieren
					XEvent event = trace.get(e);
					String eventName = xconcept.extractName(event);
					String lifecycle = xlifecycle.extractTransition(event);

					if (eventName.equals(label))
					{
						if (lifecycle.equals("start"))
						{
							startFound = true;
							selfIndep = true;
						}
						else if (lifecycle.equals("complete"))
						{
							startFound = false;
						}
					}

					// Wenn Start gefunden wurde, dann alle weiteren Events unabhängig machen
					if (startFound && !selfIndep)
					{
						independencies.add(new Tuple(label, eventName));
					}
					else
					{
						selfIndep = false;
					}
				}
			}
		}

		// Independencies ausgeben
		for (Tuple t : independencies)
		{
			//System.out.println("Unabhängigkeit: " + t.eventA + " | " + t.eventB);
		}


		//System.out.println("STOP");


		// =====================================================================
		// Prime Event Structure bauen

		PrimeEventStructure pes = new PrimeEventStructure();

		Event startEvent = new Event("initialEvent");
		pes.events.add(startEvent);
		actualEvents = new ArrayList<Event>(); // Menge der aktuell ausgewählten Events
		coSetEvents = new ArrayList<Event>(); // Menge der Events im aktuellen CoSet

		int coSetIndex = 0; // Der Index des aktuellen CoSets;
		int male = 0;

		// Speichert das CoSet vom zuletzt besuchten Event
		int letztesCoSet = 0;
		ArrayList<ArrayList<Event>> lpos = new ArrayList<ArrayList<Event>>();

		for (int t = 0; t < log.size(); t++)
		{
			try
			{
				XTrace trace = log.get(t);

				actualEvents.clear();
				actualEvents.add(startEvent);

				// Alle besuchten Events zurücksetzen
				for (Event event : pes.events)
					event.besucht = false;
				startEvent.besucht = true;

				// Wurde ein neues Event gefunden? Wenn nicht, dann muss kein
				// neues CoSet erzeugt werden
				boolean neuesCoSetErstellen = false;

				// Aktuelles CoSet zurücksetzen
				coSetEvents.clear();
				coSetEvents.add(startEvent);

				for (int e = 0; e < trace.size(); e++)
				{
					// Event auslesen
					XEvent event = trace.get(e);
					String eventName = xconcept.extractName(event);
					String lifecycle = xlifecycle.extractTransition(event);

					if (lifecycle.equals("complete"))
					{

						Object[] returnValue = istInMengeVonNachfolgern(actualEvents, eventName);
						ArrayList<Event> verbindeMenge = (ArrayList<Event>) returnValue[0];
						Event nachfolgerEvent = (Event) returnValue[1];
						boolean eventBereitsVorhanden = (boolean)returnValue[2];

						//if (vorgängerEvents != null) {
							if (eventBereitsVorhanden) {
								// Element wurde im Nachbereich der Menge von aktuellen Events gefunden
								// Man kann einfach weiterlaufen. Dafür muss die Menge der aktuellen
								// Events nur aktualisiert werden.

								// Das neue Event wird der Menge hinzugefügt. Aber nur dann, wenn der gesamte
								// Vorbereich des Events bereits besucht wurde, sonst kann dieses Event nicht
								// besucht werden
								if (alleVorgängerBesucht(nachfolgerEvent)) {
									actualEvents.add(nachfolgerEvent);
									nachfolgerEvent.besucht = true;

									// CoSetMenge wird erweitert
									coSetEvents.add(nachfolgerEvent);
								}

								// Nur wenn alle NachfolgeEvents bereits gesucht werden wird das Vorgänger
								// Event aus der Menge der aktuellen Events gelöscht
								for (Event k : verbindeMenge) {
									actualEvents.remove(k);
								}

								letztesCoSet = nachfolgerEvent.coSets.get(0);

							} else {
								// Das neue Event wurde in der bisherigen PES nicht gefunden oder es ist drin aber
								// unabhängig vom aktuellen Event.

								// Vorgänger Events mit neuem Event verbinden
								verbindeMengeVonEventsMitEvent(verbindeMenge, nachfolgerEvent);

								// Wenn es wirklich gar nicht drin ist wird ein neues CoSet wird erstellt und der PES hinzugefügt.
								neuesCoSetErstellen = true;

								// CoSet Menge wird erweitert
								coSetEvents.add(nachfolgerEvent);
								nachfolgerEvent.besucht = true;

								// Da wird ein neues CoSet angefangen haben, löschen wir alle aktuellen
								// Events und packen nur das neu erstellte Event in die Menge der aktuellen,
								// da es von diesem neuen Event jetzt weitergeht
								// Wenn die verbindeMenge gleich der actualEvents-Menge ist (d.h. man musste weiter vorne
								// keine abhängigen Events suchen, dann wird actualEvents hier nicht geleert

								if (verbindeMenge.equals(actualEvents))
									actualEvents.clear();

								// Alle besuchten Vorgänger löschen
								for (Event ev : nachfolgerEvent.previousEvents) {
									if (ev.besucht)
										actualEvents.remove(ev);
								}

								actualEvents.add(nachfolgerEvent);
							}
					}
				}

				// Neues CoSet für den aktuellen Trace soll gebaut werden
				if (neuesCoSetErstellen) {
					coSetIndex++;

					coSetCount.add(1);

					ArrayList<Event> ev = new ArrayList<Event>();
					for (Event e : coSetEvents)
					{
						ev.add(e);
					}
					lpos.add(ev);

					// Jedem Event des CoSets wird seine Index-Nummer hinzugefügt
					for (Event e : coSetEvents) {
						e.coSets.add(coSetIndex);
					}

					// Das neue CoSet bzw. die Menge der Events des CoSets der PES hinzufügen
					// Ein Event wird nur hinzugefügt, wenn es noch nicht vorhanden ist
					for (Event e : coSetEvents) {
						if (!pes.events.contains(e))
							pes.events.add(e);
					}
				}
				else
				{
					// Keine neue LPO wurde erstellt. Heißt sie ist schon in der PES
					// vorhanden. Zähler für diese LPO hochzählen (für dieses CoSet)
					coSetCount.set(letztesCoSet-1, coSetCount.get(letztesCoSet-1) + 1);
				}
			}
			catch (Exception e)
			{
				male++;
				if (debug) System.out.println("Fehler geworfen. Trace wird ignoriert. " + male + " Mal");
			}

		}

		return pes;
	}



	public static PrimeEventStructure minePES(XLog log) 
	{
		// 1=unrelated, 2=parallel, 3=causal right, 4=causal left
		int[][] dependencyMatrix;
		int matrixIndex = 0;

		// Map, mapping event names to matrix positions
		Map<String, Integer> eventMap = new HashMap<String, Integer>();
		Map<Integer, String> eventMapReverse = new HashMap<Integer, String>();

		// Dependency Matrix initialisieren

		// Number of Events holen und Traces mit mehrfachen Events löschen (Nur für die Berechnung der Unabhängigkeiten)
		ArrayList<String> labels = new ArrayList<String>();
		ArrayList<XTrace> logTraces = new ArrayList<XTrace>();

		for (int t = 0; t < log.size(); t++) {
			XTrace trace = log.get(t);

			ArrayList<String> traceLabels = new ArrayList<String>();
			boolean traceIsValid = true;

			for (int e = 0; e < trace.size(); e++) {
				XEvent eventA = trace.get(e);
				String eventNameA = xconcept.extractName(eventA);
				String lifecycle = xlifecycle.extractTransition(eventA);

				if (lifecycle != null)
					eventNameA = eventNameA + lifecycle;

				if (!traceLabels.contains(eventNameA)) {
					traceLabels.add(eventNameA);
				} else {
					traceIsValid = false;
				}

				if (!labels.contains(eventNameA))
					labels.add(eventNameA);
			}

			if (traceIsValid)
				logTraces.add(trace);

		}

		int numberOfEvents = labels.size();

		dependencyMatrix = new int[numberOfEvents][numberOfEvents];

		for (int i = 0; i < numberOfEvents; i++) {
			for (int j = 0; j < numberOfEvents; j++) {
				dependencyMatrix[i][j] = 0;
			}
		}

		// Iteriere über Traces
		for (XTrace trace : logTraces) {
			// Iteriere über Eventpaare des Trace
			for (int e = 0; e < trace.size() - 1; e++) {
				XEvent eventA = trace.get(e);
				XEvent eventB = trace.get(e + 1);

				// Extract concept:name from events
				String eventNameA = xconcept.extractName(eventA);
				String lifecycle = xlifecycle.extractTransition(eventA);

				if (lifecycle != null)
					eventNameA = eventNameA + lifecycle;

				String eventNameB = xconcept.extractName(eventB);
				lifecycle = xlifecycle.extractTransition(eventB);

				if (lifecycle != null)
					eventNameB = eventNameB + lifecycle;

				// EventMap aktualisieren
				if (!eventMap.containsKey(eventNameA)) {
					eventMap.put(eventNameA, matrixIndex);
					eventMapReverse.put(matrixIndex, eventNameA);
					matrixIndex++;
				}

				if (!eventMap.containsKey(eventNameB)) {
					eventMap.put(eventNameB, matrixIndex);
					eventMapReverse.put(matrixIndex, eventNameB);
					matrixIndex++;
				}

				// In die Matrix eintragen
				dependencyMatrix[eventMap.get(eventNameA)][eventMap.get(eventNameB)] = 1;
			}
		}
		//
		//        // Matrix ausgeben
		//        for (int i = 0; i < numberOfEvents; i++) {
		//            System.out.println("\n");
		//
		//            for (int j = 0; j < numberOfEvents; j++) {
		//                System.out.print(dependencyMatrix[i][j] + " ");
		//            }
		//        }

		// Unabhängigkeitssets erstellen
		independencies = new ArrayList<Tuple>();

		for (int i = 0; i < numberOfEvents; i++) {
			for (int j = 0; j < numberOfEvents; j++) {
				if ((dependencyMatrix[i][j] == 1) && (dependencyMatrix[j][i] == 1)) {
					// Events sind unabhängig, Events holen
					String eventA = eventMapReverse.get(i);
					String eventB = eventMapReverse.get(j);

					// Gleiche Events sind egal
					if (!eventA.equals(eventB))
						independencies.add(new Tuple(eventA, eventB));
				}
			}
		}


		// Unabhängigkeiten ausgeben
		//System.out.println("\n\nIndependencies:\n");
		for (Tuple tuple : independencies) {
			//System.out.println(tuple);
		}


		// =====================================================================
		// Prime Event Structure bauen

		PrimeEventStructure pes = new PrimeEventStructure();

		Event startEvent = new Event("initialEvent");
		pes.events.add(startEvent);
		actualEvents = new ArrayList<Event>(); // Menge der aktuell ausgewählten Events
		coSetEvents = new ArrayList<Event>(); // Menge der Events im aktuellen CoSet

		int coSetIndex = 0; // Der Index des aktuellen CoSets;

		// Speichert das CoSet vom zuletzt besuchten Event
		int letztesCoSet = 0;

		for (int t = 0; t < log.size(); t++) {
			XTrace trace = log.get(t);

			actualEvents.clear();
			actualEvents.add(startEvent);

			// Alle besuchten Events zurücksetzen
			for (Event event : pes.events)
				event.besucht = false;
			startEvent.besucht = true;

			// Wurde ein neues Event gefunden? Wenn nicht, dann muss kein
			// neues CoSet erzeugt werden
			boolean neuesCoSetErstellen = false;

			// Aktuelles CoSet zurücksetzen
			coSetEvents.clear();
			coSetEvents.add(startEvent);

			for (int e = 0; e < trace.size(); e++) {
				// Event auslesen
				XEvent event = trace.get(e);
				String eventName = xconcept.extractName(event);

				//if (eventName.equals("b"))
				//System.out.println("Hier");

				String lifecycle = xlifecycle.extractTransition(event);


				// Ist das aktuelle Event bereits in der Nachfolgermenge der actualEvents?
				Object[] returnValue = istInMengeVonNachfolgern(actualEvents, eventName);
				ArrayList<Event> verbindeMenge = (ArrayList<Event>) returnValue[0];
				Event nachfolgerEvent = (Event) returnValue[1];
				boolean eventBereitsVorhanden = (boolean)returnValue[2];


				//if (vorgängerEvents != null) {
				if (eventBereitsVorhanden) {
					// Element wurde im Nachbereich der Menge von aktuellen Events gefunden
					// Man kann einfach weiterlaufen. Dafür muss die Menge der aktuellen
					// Events nur aktualisiert werden.

					// Das neue Event wird der Menge hinzugefügt. Aber nur dann, wenn der gesamte
					// Vorbereich des Events bereits besucht wurde, sonst kann dieses Event nicht
					// besucht werden
					if (alleVorgängerBesucht(nachfolgerEvent)) {
						actualEvents.add(nachfolgerEvent);
						nachfolgerEvent.besucht = true;

						// CoSetMenge wird erweitert
						coSetEvents.add(nachfolgerEvent);
					}

					// Nur wenn alle NachfolgeEvents bereits gesucht werden wird das Vorgänger
					// Event aus der Menge der aktuellen Events gelöscht
					for (Event k : verbindeMenge) {
						actualEvents.remove(k);
					}

					letztesCoSet = nachfolgerEvent.coSets.get(0);

				} else {
					// Das neue Event wurde in der bisherigen PES nicht gefunden oder es ist drin aber
					// unabhängig vom aktuellen Event.

					// Vorgänger Events mit neuem Event verbinden
					verbindeMengeVonEventsMitEvent(verbindeMenge, nachfolgerEvent);

					// Wenn es wirklich gar nicht drin ist wird ein neues CoSet wird erstellt und der PES hinzugefügt.
					neuesCoSetErstellen = true;

					// CoSet Menge wird erweitert
					coSetEvents.add(nachfolgerEvent);
					nachfolgerEvent.besucht = true;

					// Da wird ein neues CoSet angefangen haben, löschen wir alle aktuellen
					// Events und packen nur das neu erstellte Event in die Menge der aktuellen,
					// da es von diesem neuen Event jetzt weitergeht
					// Wenn die verbindeMenge gleich der actualEvents-Menge ist (d.h. man musste weiter vorne
					// keine abhängigen Events suchen, dann wird actualEvents hier nicht geleert

					if (verbindeMenge.equals(actualEvents))
						actualEvents.clear();

					// Alle besuchten Vorgänger löschen
					for (Event ev : nachfolgerEvent.previousEvents) {
						if (ev.besucht)
							actualEvents.remove(ev);
					}

					actualEvents.add(nachfolgerEvent);
				}
			}



			// Neues CoSet für den aktuellen Trace soll gebaut werden
			if (neuesCoSetErstellen) {
				coSetIndex++;

				coSetCount.add(1);

				// Jedem Event des CoSets wird seine Index-Nummer hinzugefügt
				for (Event e : coSetEvents) {
					e.coSets.add(coSetIndex);
				}

				// Das neue CoSet bzw. die Menge der Events des CoSets der PES hinzufügen
				// Ein Event wird nur hinzugefügt, wenn es noch nicht vorhanden ist
				for (Event e : coSetEvents) {
					if (!pes.events.contains(e))
						pes.events.add(e);
				}
			}
			else
			{
				// Keine neue LPO wurde erstellt. Heißt sie ist schon in der PES
				// vorhanden. Zähler für diese LPO hochzählen (für dieses CoSet)
				coSetCount.set(letztesCoSet-1, coSetCount.get(letztesCoSet-1) + 1);
			}
		}



		return pes;
	}

	public static Object copy(Object orig) {
		Object obj = null;
		try {
			// Write the object out to a byte array
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
			ObjectOutputStream out = new ObjectOutputStream(bos);
			out.writeObject(orig);
			out.flush();
			out.close();

			// Make an input stream from the byte array and read
			// a copy of the object back in.
			ObjectInputStream in = new ObjectInputStream(
					new ByteArrayInputStream(bos.toByteArray()));
			obj = in.readObject();
		}
		catch(IOException e) {
			e.printStackTrace();
		}
		catch(ClassNotFoundException cnfe) {
			cnfe.printStackTrace();
		}
		return obj;
	}

	public static boolean eventsAreUnrelated(String eventA, String eventB)
	{
		boolean unrelated = false;

		for (Tuple tuple : independencies)
		{
			if ((tuple.eventA.equals(eventA) && tuple.eventB.equals(eventB)) ||
					(tuple.eventA.equals(eventB) && tuple.eventB.equals(eventA)))
			{
				unrelated = true;
				break;
			}
		}

		return unrelated;
	}

	public static ArrayList<Event> passendeMengeFinden(ArrayList<Event> menge, String eventName)
	{

		ArrayList<Event> passendeMenge = new ArrayList<Event>();

		int size = menge.size();
		for (int i = 0; i < size; i++)
		{
			Event event = menge.get(i);

			if (eventsAreUnrelated(event.eventName, eventName))
			{
				// Wenn beide Events unrelated sind, dann muss weiter vorne gesucht werden
				// Dafür wird für das aktuelle unrelated Event der Vorbereich zur Menge hinzugefügt
				// und das Element selber aus der Menge gelöscht
				menge.remove(i);
				size--;
				i--;

				for (Event k : event.previousEvents)
				{
					// Wir sind jetzt bei einem Event weiter nach vorne gegangen
					// Nun wird geprüft, ob das neu hinzugenommene Event bereits im gesamten Vorbereich
					// der anderen Events in menge enthalten ist. Wenn ja, können wir das einfach außen
					// vor lassen. Wenn nicht, dann wird es hinzugefügt
					// => Verhindern von transitiven Kanten in der PES

					// Also nur, wenn k nicht im Vorbereich der anderen Events in menge ist, dann hinzu
					boolean istBereitsInVorbereich = false;

					for (Event eee : menge)
					{
						if (eee.allPreviousEvents.contains(k))
							istBereitsInVorbereich = true;
					}

					if (!istBereitsInVorbereich)
					{
						menge.add(k);
						size++;
					}
				}


			}
		}

		// Wenn alle Events der Menge nicht unabhängig vom aktuellen Event e sind, dann
		// kann diese Menge zurückgegeben werden
		passendeMenge.addAll(menge);

		return passendeMenge;
	}


	//    public static Object[] istInMengeVonNachfolgern(String testEvent, ArrayList<Event> actualEvents)
	//    {
	//        ArrayList<Event> vorgängerEvents = null;
	//        Event nachfolgerEvent = null;
	//        boolean gefunden = false;
	//
	//        for (Event e : actualEvents)
	//        {
	//            for (Event ee : e.nextEvents)
	//            {
	//                if (ee.eventName.equals(testEvent))
	//                {
	//                    gefunden = true;
	//                    vorgängerEvents = new ArrayList<Event>();
	//                    vorgängerEvents.addAll(ee.previousEvents);
	//                    nachfolgerEvent = ee;
	//                    break;
	//                }
	//            }
	//
	//            if (gefunden) break;
	//        }
	//
	//        Object[] o = new Object[2];
	//        o[0] = vorgängerEvents;
	//        o[1] = nachfolgerEvent;
	//
	//        return o;
	//    }

	public static boolean alleNachfolgerBesucht(Event event)
	{
		boolean alleNachfolgerBesucht = true;

		for (Event e : event.nextEvents)
		{
			if (!e.besucht) alleNachfolgerBesucht = false;
		}

		return alleNachfolgerBesucht;
	}

	public static boolean alleVorgängerBesucht(Event event)
	{
		boolean alleVorgängerBesucht = true;

		for (Event e : event.previousEvents)
		{
			if (!e.besucht) alleVorgängerBesucht = false;
		}

		return alleVorgängerBesucht;
	}

	public static Object[] istInMengeVonNachfolgern(ArrayList<Event> actualEvents, String newEventName)
	{
		ArrayList<Event> verbindeMenge = new ArrayList<Event>(actualEvents);
		verbindeMenge = passendeMengeFinden(verbindeMenge, newEventName);

		ArrayList<Event> schnittmenge = new ArrayList<Event>();
		ArrayList<Event> vor = new ArrayList<Event>();

		for (Event nachfolger : verbindeMenge.get(0).nextEvents)
		{
			schnittmenge.add(nachfolger);
		}

		ArrayList<Event> deleteFromSchnitt = new ArrayList<Event>();

		for (Event schnittElement : schnittmenge)
		{
			for (Event e : verbindeMenge)
			{
				if (!e.nextEvents.contains(schnittElement))
				{
					deleteFromSchnitt.add(schnittElement);
				}
				else
				{
					if (!vor.contains(e)) vor.add(e);
				}
			}
		}

		schnittmenge.removeAll(deleteFromSchnitt);

		Event foundE = null;
		for (Event e : schnittmenge)
		{
			if (e.eventName.equals(newEventName))
				foundE = e;
		}

		boolean eventGefunden = false;



		if (foundE != null)
		{
			// Event wurde gefunden, ist also bereits vorhanden
			eventGefunden = true;
		}
		else
		{
			// Event wurde nicht gefunden und muss neu erzeugt werden.
			// Gleichzeitig wird auch die verbinde-Menge berechnen
			foundE = new Event(newEventName);
		}


		Object[] o = new Object[3];
		o[0] = verbindeMenge;
		o[1] = foundE;
		o[2] = eventGefunden;

		return o;
	}

	public static boolean verbindeMengeVonEventsMitEvent(ArrayList<Event> events, Event newEvent)
	{
		boolean allesGleichGeblieben = true;

		//Object o[] = istInMengeVonNachfolgern(events, newEvent.eventName);

		// Element wurde gefunden
		//        if (o[1] != null)
		//        {
		//            Event foundE = (Event)o[1];
		//            actualEvents.add(foundE);
		//            // CoSetMenge wird erweitert
		//            coSetEvents.add(foundE);
		//            foundE.besucht = true;
		//        }
		//        else
		//        {
		// Kantenverbinden
		for (Event e : events)
		{
			e.nextEvents.add(newEvent);
			newEvent.previousEvents.add(e);
			newEvent.allPreviousEvents.addAll(e.allPreviousEvents);
			newEvent.allPreviousEvents.add(e);
			allesGleichGeblieben = false;
		}
		//        }

		return allesGleichGeblieben;
	}

	// Noch kein NextState ausgewählt in der Region für das Event in dem CoSet
	public static boolean nochKeinNextStateSelected(Event e, int coSet, Region r)
	{
		boolean keinerSelected = true;


		// Iterieren über States der Region
		for (State state : r.elements) {
			if (state.inEvent.equals(e))
			{
				if (state.selected.get(coSet) != null)
				{
					if (state.selected.get(coSet) == true) keinerSelected = false;
				}
			}
		}

		return keinerSelected;
	}

	// Noch kein PreviousState ausgewählt in der Region für das Event in dem CoSet
	public static boolean nochKeinPreviousStateSelected(Event e, int coSet, Region r)
	{
		boolean keinerSelected = true;


		// Iterieren über States der Region
		for (State state : r.elements) {
			if (state.outEvent.equals(e))
			{
				if (state.selected.get(coSet) != null)
				{
					if (state.selected.get(coSet) == true) keinerSelected = false;
				}
			}
		}

		return keinerSelected;
	}
}

class Region
{
	public ArrayList<State> elements = new ArrayList<State>();

	public String toString()
	{
		String returnString = "";

		for (State s : elements)
		{
			returnString += s.toString();
		}

		return returnString;
	}
}

class State implements Serializable
{
	public Event inEvent;
	public Event outEvent;

	public Map<Integer, Boolean> markiert = new HashMap<Integer, Boolean>();
	public Map<Integer, Boolean> selected = new HashMap<Integer, Boolean>();


	public String toString()
	{
		String returnString = "";

		returnString = inEvent.eventName + "(" + inEvent.coSets.toString() + ")";
		returnString += " -> ";
		returnString += outEvent.eventName + "(" + outEvent.coSets.toString() + ")";

		return returnString;
	}
}

class PrimeEventStructure implements Serializable
{
	/**
	 *
	 */
	private static final long serialVersionUID = -5900064572282614424L;
	public ArrayList<Event> events = new ArrayList<Event>();
	public ArrayList<State> states = new ArrayList<State>();
}

class Event implements Serializable
{
	public String eventName;
	public ArrayList<Integer> coSets = new ArrayList<Integer>();

	// Speichert ALLE Vorgängerevents des Events
	public ArrayList<Event> allPreviousEvents = new ArrayList<Event>();

	boolean besucht = false;

	public ArrayList<Event> nextEvents = new ArrayList<Event>();
	public ArrayList<Event> previousEvents = new ArrayList<Event>();

	public ArrayList<State> nextStates = new ArrayList<State>();
	public ArrayList<State> previousStates = new ArrayList<State>();

	public Event(String name)
	{
		this.eventName = name;
	}

	public String toString()
	{
		return this.eventName;
	}
}

class LPO
{
	int coSet = 0;
	ArrayList<Event> events = new ArrayList<Event>();
	int count = 0;
}

class Tuple
{
	public String eventA;
	public String eventB;

	public Tuple(String eventA, String eventB)
	{
		this.eventA = eventA;
		this.eventB = eventB;
	}

	public String toString()
	{
		return "(" + eventA + " | " + eventB + ")";
	}
}

class Hasse
{
	ArrayList<Node> events;
	Node init;
	int anz;

	Hasse() 
	{
		events = new ArrayList<Node>();
	}

	Hasse(ArrayList<String> wort, HashMap<String, ArrayList<String>> conc, int anz) 
	{
		this.anz= anz;

		events = new ArrayList<Node>();

		// Sequenz bauen
		for (String s:wort)
		{
			s=s.trim();
			if (s.endsWith("complete") || s.endsWith("null"))
			{

				Node newNode=new Node(s.split("_")[0]);
				events.add(newNode);
				for(Node n:events)
				{
					n.post.add(newNode);
					newNode.pre.add(n);
				}
			}
		}

		// Unabhaeniglkeiten weghauen
		for(Node n : events)
		{

			ArrayList<Node> pre2=new ArrayList<Node>();
			for(Node m : n.pre)
			{
				if (!conc.get(n.name).contains(m.name)) 
				{
					if (!(m).equals(n)) pre2.add(m);
				}
			}   
			n.pre=pre2; 

			ArrayList<Node> post2=new ArrayList<Node>();
			for(Node m : n.post)
			{
				if (!conc.get(n.name).contains(m.name)) 
				{
					if (!(m).equals(n)) post2.add(m);
				}
			}
			n.post=post2; 
		}

		// Hasse Diagram berechnen
		for(Node i : events)
		{
			for(Node ii : events)
			{
				for(Node iii : events)
				{
					if (i.post.contains(ii) && i.post.contains(iii) && ii.post.contains(iii))
					{
						i.post.remove(iii);
						iii.pre.remove(i);
					}
				}
			}
		}
	}

	void prt()
	{
		System.out.print(anz + " : ");
		for(Node n:events)
		{
			System.out.print(n.name + " ");
		}
		System.out.println();
	}

	boolean isTheSame(Hasse h2)
	{
		// noch nicht ganz fertig! OHNE AUTO-CONCURRENCY!

		boolean passt=true;

		if (events.size() != h2.events.size()) return false;

		for (Node n:events)
		{
			String name = n.name;
			int a = 0;
			int b = 0;
			for (Node n2:events)
			{
				if (n2.name.equals(name)) a++;
			}
			for (Node n2:h2.events)
			{
				if (n2.name.equals(name)) b++;
			}
			if (a!=b) return false;
		}

		ArrayList<Node> set=new ArrayList<Node>();
		ArrayList<Node> set2=new ArrayList<Node>();

		for (Node n:events) if (n.pre.isEmpty()) set.add(n);
		for (Node n:h2.events) if (n.pre.isEmpty()) set2.add(n);

		while (!set.isEmpty())
		{
			Node n = set.get(0);
			Node n2 = null;

			for (Node x:set2) if (n.name.equals(x.name)) n2 = x;

			if (n2 == null) return false;

			set.remove(0);
			set2.remove(set2.indexOf(n2));

			set.addAll(n.post);
			set2.addAll(n2.post);
		}

		return passt;
	}
}

class Node
{
	String name;
	ArrayList<Node> pre = new ArrayList<Node>();
	ArrayList<Node> post = new ArrayList<Node>();

	Node(){}

	Node(String s)
	{
		name = s;
	}
}


class Pes
{
	ArrayList<PesNode> nodes;
	ArrayList<PesArc> arcs;

	Pes()
	{
		nodes = new ArrayList<PesNode>();
		arcs = new ArrayList<PesArc>();
	}

	void addArc(PesNode n1, PesNode n2)
	{
		PesArc arc = new PesArc(n1,n2);
		arcs.add(arc);
		n1.post.add(arc);
		n2.pre.add(arc);

	}

	void draw()
	{
		for (PesNode x:nodes) 
		{
			System.out.print("("+x.name+" ");
			for (int i: x.coset) System.out.print(i+" ");
			System.out.println(") ");
		}
		System.out.println();
		for (PesArc x:arcs) 
		{
			System.out.print("("+x.von.name+";"+x.zu.name+") ");
			System.out.println();
		}

	}

	void crop()
	{
		PesNode ini=null;
		for (PesNode n :nodes) if (n.getPreNodes().isEmpty()) ini = n; 
		crop(ini);
	}

	void crop(PesNode n)
	{
		for (int ii = n.getPostNodes().size()-1; ii >= 0; ii--)
		{
			for (int i = ii-1; i >= 0; i--)
			{
				PesNode n1=n.getPostNodes().get(i);
				PesNode n2=n.getPostNodes().get(ii);
				if (n1.name.equals(n2.name))
				{
					// gleiche Vergangenheit?
					boolean schmelzen=true;
					for (PesNode node : n1.getPreNodes())
					{
						if (!n2.getPreNodes().contains(node)) schmelzen=false;
					}
					for (PesNode node : n2.getPreNodes())
					{
						if (!n1.getPreNodes().contains(node)) schmelzen=false;
					}

					if (schmelzen)
					{
						//System.out.println("Verschmelze: "+ n1.name + " + " + n2.name);
						//verschmelze n1 und n2

						for(PesArc a : n2.post) 
						{
							a.von = n1;
							n1.post.add(a);
						}

						for(PesArc a : n2.pre) 
						{
							arcs.remove(a);
							a.von.post.remove(a);
						}

						for (int j : n2.coset) if (!n1.coset.contains(j)) n1.coset.add(j);

						nodes.remove(n2);
						i=0;
					}

				}

			}

		}

		for (PesNode x : n.getPostNodes()) crop(x);
	}

	PrimeEventStructure getPrime()
	{
		PrimeEventStructure neuPes = new PrimeEventStructure();

		HashMap<PesNode, Event> map = new HashMap<PesNode, Event>();

		PesNode ini=null;
		for (PesNode n :nodes) if (n.getPreNodes().isEmpty()) ini = n; 
		ini.name= "initialEvent";


		for (PesNode n: nodes)
		{
			Event e =new Event(n.name);
			map.put(n,e);
			neuPes.events.add(e);
			e.coSets.addAll(n.coset);
		}

		for (PesNode n: nodes)
		{
			for (PesNode p : n.getPreNodes()) map.get(n).previousEvents.add(map.get(p));
			for (PesNode p : n.getPostNodes()) map.get(n).nextEvents.add(map.get(p));
		}
		// all previousEvents?
		return neuPes;
	}
}

class PesNode
{
	ArrayList<Integer> coset= new ArrayList<Integer>();
	String name;
	ArrayList<PesArc> pre = new ArrayList<PesArc>();
	ArrayList<PesArc> post = new ArrayList<PesArc>();

	PesNode(){}

	PesNode(String s)
	{
		name = s;
	}

	ArrayList<PesNode> getPreNodes()
	{
		ArrayList<PesNode> nodes = new ArrayList<PesNode>(); 
		for (PesArc arc : pre) nodes.add(arc.von);
		return nodes; 
	}

	ArrayList<PesNode> getPostNodes()
	{
		ArrayList<PesNode> nodes = new ArrayList<PesNode>(); 
		for (PesArc arc : post) nodes.add(arc.zu);
		return nodes; 
	}
}

class PesArc
{
	PesNode von;
	PesNode zu;

	PesArc(){}

	PesArc(PesNode von, PesNode zu)
	{
		this.von = von;
		this.zu = zu;
	}
}
